/* just some information about this file */
#define DOUBLEARRAY_VERSION	"1.1.0"
#define DOUBLEARRAY_DATE	"23-Jun-2003"

/* structures and typedefs */

/** @file
* @anchor DoubleArray
* Fault-tolerant structure to store one or more double values.
*
* This structure provides a fault-tolerant way to store several double values. The individual values
* are accessed via their index @b starting @b from @b zero. Therefore, the index of the element with
* the highest index number is one less than @c numberOfElements. The memory allocation of the structure
* is managed via the @ref DoubleArray_create and @ref DoubleArray_destroy functions.
*
* The structure is created with @ref DoubleArray_create, destroyed with @ref DoubleArray_destroy and
* displayed with @ref DoubleArray_display. The structure could also be displayed with function
* @ref DoubleArray_displayIntegers; the individual values of the elements are than treated as integers.
* It could be copied with the function @ref DoubleArray_copy. The number of elements of a structure
* could be found out via @ref DoubleArray_getNumberOfElements. Elements could be assigned values with
* @ref DoubleArray_setValue and the value of an element could be obtained with @ref DoubleArray_getValue.
*
* If an element is accessed via @ref DoubleArray_setValue or @ref DoubleArray_getValue and the index
* is higher than @c numberOfElements - 1 (maximum index of this structure) the define @ref
* IndexOutOfBoundsError is called and the program aborts.
*
* <B>Description of the members:</B><BR><BR>
* @c value - the double values of the structure<BR>
* @c numberOfElements - number of elements in the structure
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
* @version 05/20/2003 - Uli Fechner - added function @ref DoubleArray_calculateMeanAndSd
* @version 05/22/2003 - Uli Fechner - added the member @c identifier und the functions @ref
* DoubleArray_createWithIdentifier and DoubleArray_getIdentifier; adjusted the function @ref
* DoubleArray_copy accordingly
* @version 05/23/2003 - Uli Fechner - SPECIAL MODIFICATION for the RankIt program: added function
* @ref DoubleArray_lessThan
* @version 06/16/2003 - v1.0.1 - Uli Fechner - SPECIAL MODIFICATION for the RankIt program: several
* adjustments in function @ref DoubleArray_calculateMeanAndSd
* @version 06/23/2003 - v1.1.0 - Uli Fechner - bugfix in function @ref DoubleArray_calculateMeanAndSd
* @@code
*/
typedef struct
{
	double* value;
	int numberOfElements;
	int identifier;
} DoubleArray;
/** @endcode */


/** A pointer to structure @ref DoubleArray is assigned the name @c DoubleArrayPtr. */
typedef DoubleArray* DoubleArrayPtr;


/* function prototypes */

DoubleArrayPtr DoubleArray_create( const int numberOfElements );

DoubleArrayPtr DoubleArray_createWithIdentifier( const int numberOfElements, const int identifier );

void DoubleArray_destroy( DoubleArrayPtr dAPtr );

void DoubleArray_display( void* dAPtr, FILE* outputStream );

void DoubleArray_displayIntegers( const DoubleArrayPtr dAPtr, FILE* outputStream );

DoubleArrayPtr DoubleArray_copy( const DoubleArrayPtr dAPtr );

int DoubleArray_getIdentifier( const DoubleArrayPtr dAPtr );

int DoubleArray_getNumberOfElements( const DoubleArrayPtr dAPtr );

void DoubleArray_setValue( const DoubleArrayPtr dAPtr, const int index, const double value );

double DoubleArray_getValue( const DoubleArrayPtr dAPtr, const int index );

DoubleArrayPtr DoubleArray_calculateMeanAndSd( const DoubleArrayPtr dAPtr );

int DoubleArray_lessThan( void* dAPtr1, void* dAPtr2 );


/* functions */

/** Creates a @ref DoubleArray structure.
*
* A @ref DoubleArray structure is created. The number of elements of the structure is given as an
* argument and is not dynamic during the lifetime of a structure @ref DoubleArray.
*
* @param numberOfElements the number of elements of the newly created structure
* @retval DoubleArrayPtr pointer on the newly created structure @ref DoubleArray
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
DoubleArrayPtr DoubleArray_create( const int numberOfElements )
{
	DoubleArrayPtr dAPtr;
	
	if( !( dAPtr = calloc( 1, sizeof( DoubleArray ) ) ) )
		MemoryError( "dAPtr", "DoubleArray_create" );
	
	dAPtr->numberOfElements = numberOfElements;
	
	if( numberOfElements == 0 )
		dAPtr->value = NULL;
	
	else
	{
		if( !( dAPtr->value = calloc( numberOfElements, sizeof( double ) ) ) )
			MemoryError( "dAPtr->value", "DoubleArray_create" );
	}

	return dAPtr;
}


/** Creates a @ref DoubleArray structure with an identifier.
*
* A @ref DoubleArray structure is created. The number of elements of the structure is given as an
* argument and is not dynamic during the lifetime of a structure @ref DoubleArray. The second
* argument is an integer that serves as a unique identifier of this @ref DoubleArray. This function
* for the creation of a @ref DoubleArray is useful, if the @ref DoubleArray is part of an array of
* DoubleArrays.
*
* @param numberOfElements the number of elements of the newly created structure
* @param identifier integer that serves as a unique identifier
* @retval DoubleArrayPtr pointer on the newly created structure @ref DoubleArray
* @author Uli Fechner
* @version 05/22/2003 - Uli Fechner - initial release
*/
DoubleArrayPtr DoubleArray_createWithIdentifier( const int numberOfElements, const int identifier )
{
	DoubleArrayPtr dAPtr;
	
	if( !( dAPtr = calloc( 1, sizeof( DoubleArray ) ) ) )
		MemoryError( "dAPtr", "DoubleArray_create" );
	
	dAPtr->numberOfElements = numberOfElements;
	dAPtr->identifier = identifier;
	
	if( numberOfElements == 0 )
		dAPtr->value = NULL;
	
	else
	{
		if( !( dAPtr->value = calloc( numberOfElements, sizeof( double ) ) ) )
			MemoryError( "dAPtr->value", "DoubleArray_create" );
	}

	return dAPtr;
}


/** Destroys a structure @ref DoubleArray.
*
* The structure @ref DoubleArray the pointer @c dAPtr refers to is destroyed. All allocated memory of
* the structure is automatically freed.
*
* @param dAPtr pointer on the structure @ref DoubleArray that should be destroyed
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
void DoubleArray_destroy( DoubleArrayPtr dAPtr )
{
	if( dAPtr != NULL )
	{
		if( dAPtr->value != NULL )
			free( dAPtr->value );
		free( dAPtr );
	}
	else
	{
		fprintf( stderr, "\n\nERROR: Function 'DoubleArray_destroy' could not destroy a structure\n" );
		fprintf( stderr, "'DoubleArray' that has not been created before!\n" );
		AbortProgram;
	}
}


/** Copies a structure @ref DoubleArray.
*
* A copy of the structure @ref DoubleArray referred to by @c dAPtr is made. The copy lives in its own
* memory and contains all elements the template contains (sometimes called @c deepcopy).
*
* @param dAPtr pointer on the structure @ref DoubleArray that should be copied
* @retval DoubleArrayPtr pointer on structure @ref DoubleArray containing the copy
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
* @version 05/22/2003 - Uli Fechner - added the copying of the identifier
*/
DoubleArrayPtr DoubleArray_copy( const DoubleArrayPtr dAPtr )
{
	int counter;
	DoubleArrayPtr copyOfDAPtr;
	
	if( dAPtr == NULL )
	{
		fprintf( stderr, "\n\nERROR: Function 'DoubleArray_copy' could not copy a structure\n" );
		fprintf( stderr, "'DoubleArray' that is empty!\n" );
		AbortProgram;
	}
	
	if( !( copyOfDAPtr = calloc( 1, sizeof( DoubleArray ) ) ) )
		MemoryError( "copyOfDAPtr", "DoubleArray_copy" );
	
	copyOfDAPtr->numberOfElements = DoubleArray_getNumberOfElements( dAPtr );
	copyOfDAPtr->identifier = DoubleArray_getIdentifier( dAPtr );
	
	if( DoubleArray_getNumberOfElements( dAPtr ) == 0 )
		copyOfDAPtr->value = NULL;
	
	else
	{
		if( !( copyOfDAPtr->value = calloc( DoubleArray_getNumberOfElements( dAPtr ), sizeof( double ) ) ) )
			MemoryError( "copyOfDAPtr->value", "DoubleArray_copy" );
	}

	for( counter = 0; counter < DoubleArray_getNumberOfElements( dAPtr ); counter++ )
		DoubleArray_setValue( copyOfDAPtr, counter, DoubleArray_getValue( dAPtr, counter ) );
	
	return copyOfDAPtr;
}

/** Displays a structure @ref DoubleArray.
*
* The structure @ref DoubleArray the pointer @c dAPtr refers to is displayed on the FILE* @c
* outputStream.
*
* @param dAPtr pointer on the structure @ref DoubleArray that should be displayed
* @param outputStream FILE* on the stream the output should be sent to
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
void DoubleArray_display( void* dAPtr, FILE* outputStream )
{
	int counter; /* simple counter variable in for loop */
	
	if( dAPtr == NULL )
	{
		fprintf( stderr, "\n\nERROR: Function 'DoubleArray_display' could not display a structure\n" );
		fprintf( stderr, "'DoubleArray' that is empty!\n" );
		AbortProgram;
	}

	for( counter = 0; counter < DoubleArray_getNumberOfElements( dAPtr ); counter++ )
		fprintf( outputStream, "%.6f  ", DoubleArray_getValue( dAPtr, counter ) );
	fprintf( outputStream, "\n" ); 
}

/** Displays a structure @ref DoubleArray.
*
* The structure @ref DoubleArray the pointer @c dAPtr refers to is displayed on the FILE* @c
* outputStream. The values of each element contained in @ref DoubleArray are treated as integers.
*
* @param dAPtr pointer on the structure @ref DoubleArray that should be displayed
* @param outputStream FILE* on the stream the output should be sent to
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
void DoubleArray_displayIntegers( const DoubleArrayPtr dAPtr, FILE* outputStream )
{
	int counter; /* simple counter variable in for loop */
	
	if( dAPtr == NULL )
	{
		fprintf( stderr, "\n\nERROR: Function 'DoubleArray_display' could not display a structure\n" );
		fprintf( stderr, "'DoubleArray' that is empty!\n" );
		AbortProgram;
	}

	for( counter = 0; counter < DoubleArray_getNumberOfElements( dAPtr ); counter++ )
		fprintf( outputStream, "%d  ", (int) DoubleArray_getValue( dAPtr, counter ) );
	fprintf( outputStream, "\n" ); 
}


/** Returns the identifier of a structure @ref DoubleArray.
*
* @param dAPtr pointer on the structure @ref DoubleArray of which the number of elements should be returned
* @retval int unique identifier of the structure @ref DoubleArray
* @author Uli Fechner
* @version 05/22/2003 - Uli Fechner - initial release
*/
int DoubleArray_getIdentifier( const DoubleArrayPtr dAPtr )
{
	return dAPtr->identifier;
}

/** Returns the number of elements of a structure @ref DoubleArray.
*
* The number of elements of the structure @ref DoubleArray the pointer @c dAPtr refers to is returned.
*
* @param dAPtr pointer on the structure @ref DoubleArray of which the number of elements should be returned
* @retval int number of elements of the structure @ref DoubleArray
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
int DoubleArray_getNumberOfElements( const DoubleArrayPtr dAPtr )
{
	return dAPtr->numberOfElements;
}


/** The element with index @c index of structure @ref DoubleArray is assigned the value @c value.
*
* The value @c value is assigned to the element with index @c index of the structure @ref DoubleArray
* @c dAPtr points on. If the index @c index is greater than (@c numberOfElements - 1 ), the program
* aborts with the call of the define @ref IndexOutOfBoundsError.
*
* @param dAPtr pointer on the structure @ref DoubleArray that contains the element
* @param index index of the element the value @c value should be assigned to
* @param value value that should be assigned to the element with index @c index
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
void DoubleArray_setValue( const DoubleArrayPtr dAPtr, const int index, const double value )
{
	if( index >= DoubleArray_getNumberOfElements( dAPtr ) )
		IndexOutOfBoundsError( "DoubleArray", DoubleArray_getNumberOfElements( dAPtr ), index );

	(dAPtr->value)[ index ] = value;
}

/** The value of the element with index @c index of structure @ref DoubleArray is returned.
*
* The value of the element with index @c index of the structure @ref DoubleArray @c dAPtr points on is
* returned. If the index @c index is greater than (@c numberOfElements - 1 ), the program aborts with
* the call of the define @ref IndexOutOfBoundsError.
*
* @param dAPtr pointer on the structure @ref DoubleArray that contains the element
* @param index index of the element the value should be returned
* @retval double the value of the element with index @c index
* @author Uli Fechner
* @version 01/24/2003 - Uli Fechner - initial release
*/
double DoubleArray_getValue( const DoubleArrayPtr dAPtr, const int index )
{
	if( index < DoubleArray_getNumberOfElements( dAPtr ) )
		return (dAPtr->value)[ index ];
	else
		IndexOutOfBoundsError( "DoubleArray", DoubleArray_getNumberOfElements( dAPtr ), index );
}


/** The mean and standard deviation of the values contained in one @ref DoubleArray is calculated.
*
* The return value of this function is a DoubleArrayPtr. The corresponding @ref DoubleArray contains
* two values:
* @li 1st value - arithmetic mean
* @li 2nd value - standard deviation
*
* @param dAPtr pointer on the structure @ref DoubleArray that contains the elements
* @retval DoubleArrayPtr DoubleArray with two values: mean and standard deviation
* @author Uli Fechner
* @version 05/20/2003 - Uli Fechner - initial release
* @version 05/23/2003 - Uli Fechner - SPECIAL MODIFICATION for the RankIt program (see remarks)
* @version 06/16/2003 - Uli Fechner - SPECIAL MODIFICATION for the RankIt program (see remarks)
* @version 06/23/2003 - Uli Fechner - IMPORTANT bugifx: the standard deviation for the active compounds
* was not calculated correctly
*/
DoubleArrayPtr DoubleArray_calculateMeanAndSd( const DoubleArrayPtr dAPtr )
{
	int counter; /* simpler counter variable in for loop */
	int numberOfElements; /* number of elements of the DoubleArray */
	double mean = 0.0f; /* used to store the calculated mean value */
	double standardDeviation = 0.0f; /* used to store the calculated standard deviation */
	/* DoubleArray that contains two values: first the mean and second the standard deviation of dAPtr */
	DoubleArrayPtr dAMeanAndSd;
	/* if one value of the DoubleArray is zero the mean is divided by numberOfElements - 1 */
	int zeroContained = BOOLEAN_FALSE;
	
	for( counter = 0; counter < DoubleArray_getNumberOfElements( dAPtr ); counter++ )
		if( ( (int) DoubleArray_getValue( dAPtr, counter ) ) == 0 )
			zeroContained = BOOLEAN_TRUE;
	
	dAMeanAndSd = DoubleArray_create( 2 );
	numberOfElements = DoubleArray_getNumberOfElements( dAPtr );
		
	/* calculating the mean */
	for( counter = 0; counter < numberOfElements; counter++ )
		mean += DoubleArray_getValue( dAPtr, counter );
	/* if one element of the DoubleArray is zero the mean is divided by numberOfElements - 1.
	This is a special modification for the rankIt program. The element that is zero indicates that
	the corresponding Compound is an active compound. Every active compound is one time the query
	structure. Hence, it has no real rank for this similarity search and the mean has to be calculated
	with one element less than the other, non-active compounds. */
	mean /= numberOfElements - zeroContained;
	DoubleArray_setValue( dAMeanAndSd, 0, mean );
	
	/* if there is only one element in daPtr and this element is zero the mean is set to zero */
	if( numberOfElements - zeroContained == 0 )
		DoubleArray_setValue( dAMeanAndSd, 0, 0.0f );
	
	/* calculating the standard deviation */
	for( counter = 0; counter < numberOfElements; counter++ )
	{
		if( DoubleArray_getValue( dAPtr, counter ) != 0 )
			standardDeviation += pow( mean - DoubleArray_getValue( dAPtr, counter ), 2 );
	}
	standardDeviation /= ( numberOfElements - 1 - zeroContained );
	standardDeviation = sqrt( standardDeviation );
	DoubleArray_setValue( dAMeanAndSd, 1, standardDeviation );
	
	/* it does not make any sense to calculate a standard deviation if there are only 2 values and one
	of these two is zero, i.e. the compound is an active compound OR if there is only one value at all */
	if( ( numberOfElements == 2 && zeroContained == BOOLEAN_TRUE ) || ( numberOfElements ==1 ) )
		DoubleArray_setValue( dAMeanAndSd, 1, 0.0f );
	
	return dAMeanAndSd;
}

/** Evaluates if the @c first value of @c dAPtr1 is less than that of @c dAPtr2.
*
* The function returns @ref BOOLEAN_TRUE if the @c first value of @c dAPtr 1 is less than the
* @c first value of @c dAPtr2 and @ref BOOLEAN_FALSE if this condition is not matched.
*
* @param dAPtr1 pointer on the first structure @ref DoubleArray
* @param dAPtr2 pointer on the second structure @ref DoubleArray
* @author Uli Fechner
* @version 05/23/2003 - Uli Fechner - initial release
*/
int DoubleArray_lessThan( void* dAPtr1, void* dAPtr2 )
{
	if( DoubleArray_getValue( dAPtr1, 0 ) < DoubleArray_getValue( dAPtr2, 0 ) )
		return BOOLEAN_TRUE;
	else
		return BOOLEAN_FALSE;
}
