/** @mainpage retroflux
* <p>
* @b retroFlux facilitates the virtual retro-synthesis of compounds. The user has to specify an <i>input
* file</i> that contains compounds in SMILES notation and a <i>reaction file</i> that contains the
* reactions in SMIRKS/reaction SMILES notation. All compounds of the <i>input file</i> are exhaustively
* retro-synthesized with all reactions specified in the <i>reaction file</i>. The result is a list of
* fragments. 'Virtual atoms' are added to tag the reaction site and reaction type of the fragments.
* In SMILES notation these 'virtual atoms' look like <tt>[n*]</tt> where n is the number of the
* respective reaction.
* <p>
* This distribution comes
* together with a file called <i>reactions.smi</i> that contains eleven reactions. Both the virtual
* retro-synthesis procedure and the eleven reactions of <i>reactions.smi</i> were originally published
* in a study by Lewell and co-workers [1].
* <p>
* Furthermore, @b retroFlux allows to provide a 'filter-file' that contains <i>unwanted fragments</i> in
* SMILES notation. Reactions that lead to one of the fragments specified in the 'filter-file' are not
* carried out. If no such file is specified all resultant fragments are considered as valid.
* <p>
* <b>RetroFlux</b> was presented at EuroMUG 2004, Cambridge, England. If you wish to publish results
* obtained using <b>retroflux</b> please cite [1] and [2].
* <p>
* @b References
* <ol>
* <li>Lewell, X.Q., Judd, D.B., Watson, S.P., and Hann, M.M., RECAP - Retrosynthetic Combinatorial
* Analysis Procedure: A Powerful New Technique for Identifying Privileged Molecular Fragments with
* Useful Applications in Combinatorial Chemistry. <i>J. Chem. Inf. Comput. Sci.</i>, 38 <b>(1998)</b>,
* 511.</li>
* <li>G. Schneider, U. Fechner, Fragment-based de novo design, <i>Presentation at EuroMUG 2004</i>,
* November 5th <b>2004</b>, Cambridge, England,
* <a href="http://daylight.com/meetings/emug04/Schneider/index.html">
* http://daylight.com/meetings/emug04/Schneider/index.html</a>.</li>
* </ol>
* <p>
* &copy; 2003-2004 by Uli Fechner
*/

/** @file
*
* This is the main file of the program flux. All other files are included here via the inclusion
* of the file @ref includes.h.
*
* @author Uli Fechner
* @version 10/12/2003 - Uli Fechner - 0.1 - initial stable release
* @version 17/12/2003 - Uli Fechner - 0.2 - changed function prototype of @ref enqueueMolecules; major
* changes in @ref enqueueMolecules; adjusted corresponding function call in @ref main
* @version 07/01/2004 - UF - 0.3 - introduced the filter SMILES file; added function @ref
* filterMolecules, changes in @ref main, @ref parseClp and @ref displayHelpText
* @version 30/01/2004 - UF - 0.3.1 - small changes in @ref enqueueMolecules, @ref filterMolecules
* and @ref readDataFromStream to reflect changes in the file smilesCompound.c
* @version 23/02/2004 - UF - 0.3.2 - added a reaction counter in the stderr output (@ref main)
* @version 16/03/2004 - UF - 0.3.3 - function SmilesCompound_getDaylightHandle was renamed to @ref
* SmilesCompound_getMoleculeHandle; this change is now reflected in retroflux
* @version 06/08/2004 - UF - 0.3.4 - changes in @ref readDataFromFile and @ref readDataFromStream
*/

/* preprocessor includes */

#include "includes.h"

/* just some information about this file */
#define RETROFLUX_VERSION	"0.3.4"
#define RETROFLUX_DATE	"06-Aug-2004"

/* function prototypes */

int main( int argc, char *argv[] );

int filterMolecules( char* reactants, const List_Ptr filterSmilesPtr, FILE* errorLogFile );

void enqueueMolecules( char* reactants, char* reactionName, const SmilesCompound_Ptr product, \
	const List_Ptr listPtr, FILE* errorLogFile );

void parseClp( const int argc, char* argv[], CLP_Ptr clpPtr );

List_Ptr readDataFromFile( const char* const filename, const int daylight_type, const int uniqueData, \
	FILE* errorLogFile );

List_Ptr readDataFromStream( FILE* inputStream, const char* const nameOfStream, \
	const int daylight_type, const int uniqueData, FILE* errorLogFile );

DoubleArrayPtr getFileProperties( FILE* inputStream, const char* const nameOfStream );

void displayHelpText( const CLP_Ptr clpPtr );

void displayVersionInformation( void );

/* functions */

/** Main function of the program.
*
* The main function just calls the other functions. It is reduced to contain only function calls
* whenever possible.
*
* @param argc the number of arguments provided via the command line by executing the program
* @param *argv[] string array containing the command line arguments
* @retval int exit code of the program
* @author Uli Fechner
* @version 10/12/2003 - Uli Fechner - initial release
* @version 17/12/2003 - Uli Fechner - adjusted function call of @ref enqueueMolecules
* @version 07/01/2004 - UF - major change: introduced the filter SMILES file; this leads to several
* changes in the nested while loops as well as the addition of the reading of the filter SMILES file
* @version 23/03/2004 - UF - added a reaction counter in the stderr output
*/
int main ( int argc, char *argv[] )
{
	CLP_Ptr clpPtr; /* pointer on CommandLineParameters */
	FILE* outFile; /* file pointer on the output file */
	FILE* errorLogFile; /* file pointer on the SMILES error log file */
	List_Ptr inputDataPtr; /* pointer on a structure List that stores the data of the input file */
	List_Ptr reactionSmilesPtr; /* pointer on a List that stores the data of the reaction smiles file */
	List_Ptr filterSmilesPtr = NULL; /* pointer on a List that stores the data of the filter SMILES file */
	SmilesCompound_Ptr currentReactionPtr; /* pointer on current reaction SMILES in OUTER LOOP */
	SmilesCompound_Ptr currentSmilesPtr; /* pointer on current input SMILES in INNER LOOP */
	int totalNumberOfCompounds = 0; /* total number of compounds in input file */
	/* boolean variable indicating whether a SMILES of inputDataPtr was successfully transformed and does
	not lead to reactants that are contained in the filter SMILES file */
	int moleculeEnqueued;
	int reactionCounter; /* counts the number each reaction takes place */
	
	/***** Daylight Toolkit variables *****/
	dt_Handle errorSequenceHandle; /* dt_Handle on a stream that contains the error messages of the DT */
	dt_Handle errorHandle; /* dt_Handle on one error message of errorSequenceHandle */
	dt_Integer length; /* needed by dt_cansmiles, dt_typename and dt_stringvalue */
	/* Daylight Toolkit variables of the OUTER LOOP */
	dt_Handle reactionSmilesHandle; /* dt_Handle on current reaction SMILES */
	/* Daylight Toolkit variables of the INNER LOOP */
	dt_Handle currentSmilesHandle; /* dt_Handle on the current input SMILES */
	dt_Handle transformSequenceHandle; /* dt_Handle on current transform sequence */
	dt_Handle reactionHandle; /* dt_Handle on the reactions retrieved from transformSequenceHandle */
	dt_Handle moleculeStreamHandle; /* dt_Handle on the molecule stream retrieved from the reactionHandle */
	dt_Handle moleculeHandle; /* dt_Handle on the reactants retrieved from moleculeStreamHandle */
	
	/* creating a CommandLineParameters structure and initialize it with default values */
	clpPtr = CLP_create( );
	parseClp( argc, argv, clpPtr ); /* parsing the command line arguments */
	/* the command line arguments are shown on standard error */
	fprintf( stderr, "\nretroflux %s ( %s )\n\n", RETROFLUX_VERSION, RETROFLUX_DATE );
	CLP_display( clpPtr, stderr );
	
	/* opening the output file; if an error occurs the program aborts */
	if( !( outFile = fopen( CLP_getOutputFile( clpPtr ), "w" ) ) )
		FileWriteError( CLP_getOutputFile( clpPtr ) );
	
	/* opening the SMILES error log file; if an error occurs the program aborts */
	if( !( errorLogFile = fopen( CLP_getErrorLogFile( clpPtr ), "w" ) ) )
		FileWriteError( CLP_getErrorLogFile( clpPtr ) );

	/* reading the data from stdin */
	fprintf( stderr, "Reading file from stdin......." );
	inputDataPtr = readDataFromStream( stdin, "<stdin>", DAYLIGHT_SMILES, BOOLEAN_TRUE, errorLogFile );
	List_setName( inputDataPtr, "Database" );
	List_rewind( inputDataPtr );
	fprintf( stderr, "Done\n" );
	while( List_hasNext( inputDataPtr ) )
		totalNumberOfCompounds += SmilesCompound_getCounter( List_getNext( inputDataPtr ) );
	fprintf( stderr, "Total number of compounds in input file: %d\n", totalNumberOfCompounds );
	fprintf( stderr, "Number of unique compounds in input file: %d\n\n", \
		List_getNumberOfNodes( inputDataPtr ) );
	#if RETROFLUX_DEBUG
	List_display( inputDataPtr, outFile );
	#endif

	/* reading reaction SMILES file (provided with '-r' option) */
	fprintf( stderr, "Reading reaction SMILES file %s...", CLP_getReactionFile( clpPtr ) );
	reactionSmilesPtr = readDataFromFile( CLP_getReactionFile( clpPtr ), DAYLIGHT_SMIRKS, BOOLEAN_TRUE, \
		errorLogFile );
	List_setName( reactionSmilesPtr, "Reaction SMILES" );
	fprintf( stderr, "Done\n" );
	fprintf( stderr, "Number of reactions: %d\n", List_getNumberOfNodes( reactionSmilesPtr ) );
	#if RETROFLUX_DEBUG
	List_display( reactionSmilesPtr, outFile );
	#endif
	
	/* reading filter SMILES file (provided with '-f' option) */
	if( CLP_getFilterFile( clpPtr ) != NULL )
	{
		fprintf( stderr, "Reading filter SMILES file %s...", CLP_getFilterFile( clpPtr ) );
		filterSmilesPtr = readDataFromFile( CLP_getFilterFile( clpPtr ), DAYLIGHT_SMILES, BOOLEAN_TRUE, \
			errorLogFile );
		List_setName( filterSmilesPtr, "Filter SMILES" );
		fprintf( stderr, "Done\n" );
		fprintf( stderr, "Number of filter SMILES: %d\n", List_getNumberOfNodes( filterSmilesPtr ) );
		#if RETROFLUX_DEBUG
		List_display( filterSmilesPtr, outFile );
		#endif
	}

	/* rewinding the lists inputDataPtr and reactionSmilesPtr */
	List_rewind( inputDataPtr );
	List_rewind( reactionSmilesPtr );
	/***** OUTER LOOP: LOOPING OVER THE REACTIONS *****/
	while( List_hasNext( reactionSmilesPtr ) )
	{
		reactionCounter = 0; // re-setting the reaction counter
		currentReactionPtr = List_getNext( reactionSmilesPtr );
		fprintf( stderr, "Processing reaction %s....", \
		StringArray_getElement( SmilesCompound_getStringArrayOfNames( currentReactionPtr ), 0 ) );
		/*SmilesCompound_display( currentReactionPtr, outFile ); */
		reactionSmilesHandle = SmilesCompound_getMoleculeHandle( currentReactionPtr );
		/***** INNER LOOP: LOOPING OVER THE INPUT DATA *****/
		while( List_hasNext( inputDataPtr ) )
		{
			moleculeEnqueued = BOOLEAN_FALSE;
			currentSmilesPtr = List_getNext( inputDataPtr );
			currentSmilesHandle = SmilesCompound_getMoleculeHandle( currentSmilesPtr );
			/*SmilesCompound_display( currentSmilesPtr, outFile ); */
			transformSequenceHandle = dt_utransform( reactionSmilesHandle, currentSmilesHandle, DX_REVERSE, 1 );
			
			/* Looping over the individual REACTIONS of the transform sequence */
			while( NULL_OB != ( reactionHandle = dt_next( transformSequenceHandle ) ) )
			{
				/*fprintf( outFile, "Complete Reaction: %s\n", dt_cansmiles( &length, reactionHandle, 1 ) );*/
				/* Looping over the individual MOLECULES of one reaction */
				moleculeStreamHandle = dt_stream( reactionHandle, TYP_MOLECULE );
				/*fprintf( stderr, "%d\n", dt_count( moleculeStreamHandle, TYP_MOLECULE ) );*/
				while( NULL_OB != ( moleculeHandle = dt_next( moleculeStreamHandle ) ) )
				{
					if( dt_getrole( moleculeHandle, reactionHandle ) != DX_ROLE_REACTANT )
						break; //LOOP EXITS HERE
					/* fprintf( outFile, "%s\n%s\n", StringArray_getElement( \
						SmilesCompound_getName( currentSmilesPtr ), 0 ), \
						dt_cansmiles( &length, moleculeHandle, 1 ) ); fflush( NULL ); */
					/* if one of the reactants is contained in the filter file the reactants are not enqueued */
					if( ( CLP_getFilterFile( clpPtr ) != NULL ) && \
						( filterMolecules( dt_cansmiles( &length, moleculeHandle, 1 ), \
						filterSmilesPtr, errorLogFile ) == BOOLEAN_TRUE ) )
						break; //LOOP EXITS HERE
					enqueueMolecules( dt_cansmiles( &length, moleculeHandle, 1 ), \
						StringArray_getElement( SmilesCompound_getStringArrayOfNames( currentReactionPtr ), 0 ), \
						currentSmilesPtr, inputDataPtr, errorLogFile );
					moleculeEnqueued = BOOLEAN_TRUE;
					reactionCounter++;
					/*List_display( inputDataPtr, outFile ); fflush( NULL ); */
				}
			}
			/* the current SMILES in the input file is removed if the transform was applied successfully
			AND the reactants are not contained in the filter SMILES file */
			if( ( dt_count( transformSequenceHandle, TYP_REACTION ) ) != -1 && \
				moleculeEnqueued == BOOLEAN_TRUE )
				List_remove( inputDataPtr );
			dt_dealloc( transformSequenceHandle );
		}
		/***** END OF INNER LOOP: LOOPING OVER THE INPUT DATA *****/
		/*fprintf( outFile, "FINISHED: %s\n", \
			StringArray_getElement( SmilesCompound_getStringArrayOfNames( currentReactionPtr ), 0 ) );*/
		//List_display( inputDataPtr, outFile ); fflush( NULL ); //DEBUG
		List_rewind( inputDataPtr );
		fprintf( stderr, "%d\n", reactionCounter );
	}
	/***** END OF OUTER LOOP: LOOPING OVER THE REACTIONS *****/

	/* writing the Daylight Toolkit errors to the error log file */
	errorSequenceHandle = dt_errors( DX_ERR_NONE );
	while( NULL_OB != ( errorHandle = dt_next( errorSequenceHandle ) ) )
		fprintf( errorLogFile, "%s\n", dt_stringvalue( &length, errorHandle ) );
	
	/* writing the final file to outFile */
	List_display( inputDataPtr, outFile );
	
	/* cleaning up */
	List_destroy( inputDataPtr );
	List_destroy( reactionSmilesPtr );
	if( CLP_getFilterFile( clpPtr ) != NULL )
		List_destroy( filterSmilesPtr );
	fclose( outFile );
	CLP_destroy( clpPtr );
	
	fprintf( stderr, "Adios.\n\n");
	return EXIT_SUCCESS;
}

/** Takes a SMILES string with several molecules and checks if one of these molecules is contained in @c filterSmilesPtr.
*
* This function takes a SMILES string with several molecules that are separated by a '.' (@c 
* reactants) and stores each molecule individually in the List @c listPtr points on. Molecules
* are inserted at the tail of the list. @c reactionName and @c errorLogFile are used as a function
* argument to @ref SmilesCompound_create.
*
* @param reactants SMILES string that contains several molecules
* @param filterSmilesPtr pointer on a List that stores the molecules that should be filtered
* @param errorLogFile file pointer on the error log file
* @author Uli Fechner
* @version 07/01/2004 - Uli Fechner - initial release
*/
int filterMolecules( char* reactants, const List_Ptr filterSmilesPtr, FILE* errorLogFile )
{
	char* substring; /* used with strstr to go from '.' to '.' in reactants */
	char* copyStart; /* used together with substring to retrieve individual molecules out of reactants */
	char* tempSmiles; /* temporary storage for the SMILES contained in reactants */
	/* set to BOOLEAN_TRUE if one of the molecules of reactants is contained in filterMoleculesPtr */
	int containsFilterMolecule = BOOLEAN_FALSE;
	SmilesCompound_Ptr scPtr; /* pointer on a newly created structure SmilesCompound */
	
	if( !( tempSmiles = calloc( strlen( reactants ) + 1, sizeof( char ) ) ) )
		MemoryError( "tempSmiles", "filterMolecules" );
		
	copyStart = reactants;
	/* looping over all molecules contained in reactants (separated by a '.') */
	while( ( substring = strstr( copyStart, "." ) ) != NULL )
	{
		strncpy( tempSmiles, copyStart, (substring - copyStart ) );
		tempSmiles[ substring - copyStart ] = '\0';
		copyStart = substring + 1;
		/* insertion of the molecule at the tail of the list */
		scPtr = SmilesCompound_create( NULL, tempSmiles, DAYLIGHT_SMILES, errorLogFile );
		if( List_isContained( filterSmilesPtr, scPtr ) == BOOLEAN_TRUE )
			containsFilterMolecule = BOOLEAN_TRUE;
		SmilesCompound_destroy( scPtr );
	}
	/* copying the last molecule contained in reactants */
	strncpy( tempSmiles, copyStart, ( ( reactants + strlen( reactants ) ) - ( copyStart ) ) );
	tempSmiles[ ( ( reactants + strlen( reactants ) ) - ( copyStart ) ) ] = '\0';
	/* insertion of the last molecule at the tail of the list */
	scPtr = SmilesCompound_create( NULL, tempSmiles, DAYLIGHT_SMILES, errorLogFile );
	if( List_isContained( filterSmilesPtr, scPtr ) == BOOLEAN_TRUE )
		containsFilterMolecule = BOOLEAN_TRUE;
	SmilesCompound_destroy( scPtr );

	/* cleaning up */
	free( tempSmiles );
	
	return containsFilterMolecule;
}

/** Takes a SMILES string with several molecules and adds each molecule to @c reactants.
*
* This function takes a SMILES string with several molecules that are separated by a '.' (@c 
* reactants) and stores each molecule individually in the List @c listPtr points on. Molecules
* are inserted at the tail of the list. @c reactionName and @c errorLogFile are used as a function
* argument to @ref SmilesCompound_create.
*
* @param reactants SMILES string that contains several molecules
* @param reactionName name of the reaction the SMILES string resulted from
* @param product 
* @param listPtr pointer on a List that stores the individual molecules that result from @c reactants
* @param errorLogFile file pointer on the error log file
* @author Uli Fechner
* @version 04/12/2003 - Uli Fechner - initial release
* @version 07/12/2003 - Uli Fechner - the reactants are now named according to the scheme
* compoundName.reactionName, where compoundName is the name of the product
* @version 17/12/2003 - Uli Fechner - major changes related to the creation of the reactant
* @ref SmilesCompound
*/
void enqueueMolecules( char* reactants, char* reactionName, const SmilesCompound_Ptr product, \
	const List_Ptr listPtr, FILE* errorLogFile )
{
	char* substring; /* used with strstr to go from '.' to '.' in reactants */
	char* copyStart; /* used together with substring to retrieve individual molecules out of reactants */
	char* tempSmiles; /* temporary storage for the SMILES contained in reactants */
	char* dotPlusReactionName; /* a dot plus the name of the reaction (=.reactionName) */
	SmilesCompound_Ptr scPtr; /* pointer on a newly created structure SmilesCompound */
	
	if( !( tempSmiles = calloc( strlen( reactants ) + 1, sizeof( char ) ) ) )
		MemoryError( "tempSmiles", "enqueueMolecules" );
	if( !( dotPlusReactionName = calloc( strlen( reactionName ) + 2, sizeof( char ) ) ) )
		MemoryError( "dotPlusReactionName", "enqueueMolecules" );
	strcpy( dotPlusReactionName, "." );
	strcat( dotPlusReactionName, reactionName );
		
	copyStart = reactants;
	/* looping over all molecules contained in reactants (separated by a '.') */
	while( ( substring = strstr( copyStart, "." ) ) != NULL )
	{
		strncpy( tempSmiles, copyStart, (substring - copyStart ) );
		tempSmiles[ substring - copyStart ] = '\0';
		copyStart = substring + 1;
		/* insertion of the molecule at the tail of the list */
		scPtr = SmilesCompound_copy( product );
		SmilesCompound_setSmiles( scPtr, tempSmiles, DAYLIGHT_SMILES, errorLogFile );
		StringArray_mapConcatenate( SmilesCompound_getStringArrayOfNames( scPtr ), dotPlusReactionName );
		if( List_insertTail( listPtr, scPtr ) == BOOLEAN_FALSE )
			SmilesCompound_destroy( scPtr );
	}
	/* copying the last molecule contained in reactants */
	strncpy( tempSmiles, copyStart, ( ( reactants + strlen( reactants ) ) - ( copyStart ) ) );
	tempSmiles[ ( ( reactants + strlen( reactants ) ) - ( copyStart ) ) ] = '\0';
	/* insertion of the last molecule at the tail of the list */
	scPtr = SmilesCompound_copy( product );
	SmilesCompound_setSmiles( scPtr, tempSmiles, DAYLIGHT_SMILES, errorLogFile );
	StringArray_mapConcatenate( SmilesCompound_getStringArrayOfNames( scPtr ), dotPlusReactionName );
	if( List_insertTail( listPtr, scPtr ) == BOOLEAN_FALSE )
		SmilesCompound_destroy( scPtr );

	/* cleaning up */
	free( tempSmiles );
	free( dotPlusReactionName );
}

/** Parses the command line parameters.
*
* The command line parameters are parsed here. Additionally, it is checked whether the standard in is
* connected to a terminal. If this is the case, no input file is provided and the help text of the
* program is printed on standard error. It is also checked here if there are any incompatibilities
* of the options provided at the command line. If so, the program aborts with an appropriate error
* message.
*
* @param argc the number of arguments provided via the command line by executing the program
* @param *argv[] string array containing the command line arguments
* @param clpPtr pointer on structure @ref CommandLineParameters that is used to store all the command line parameters
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
* @version 02/12/2003 - Uli Fechner - changed option '-r' to mandatory
* @version 03/12/2003 - Uli Fechner - now automatically setting the name of the error log file
* @version 01/07/2004 - UF - added support for the f option (filter file)
*/
void parseClp( const int argc, char* argv[], CLP_Ptr clpPtr )
{	
	int c; /* gets the arguments via getopt */
	char* errorLogFile; /* temporary storage of the error log file name */
	Given_CLP_Ptr givenClpPtr; /* pointer on structure GivenCommandLineParameters */	

	givenClpPtr = GivenClp_create( ); /* creating and initializing a structure GivenCommandLineParameters */
	
	opterr = 0; /* disable the error message getopt normaly prints on stderr */
	
	while ( ( c = getopt( argc, argv, "hvo:r:f:" ) ) != -1 )
	{
		switch( c )
		{
		/* checking if to provide the help text */
		case 'h':
			displayHelpText( clpPtr );
			break;

		/* checking if to provide version information text */
		case 'v':
			displayVersionInformation();
			break;

		/* the argument of the argument o defines the name of the output file */
		case 'o':
			CLP_setOutputFile( clpPtr, optarg );
			GivenClp_setOption( givenClpPtr, 'o', BOOLEAN_TRUE );
			break;

		/* the argument of the argument o defines the name of the reaction SMILES file */
		case 'r':
			CLP_setReactionFile( clpPtr, optarg );
			GivenClp_setOption( givenClpPtr, 'r', BOOLEAN_TRUE );
			break;

		/* the argument of the argument f defines the name of the filter SMILES file */
		case 'f':
			CLP_setFilterFile( clpPtr, optarg );
			GivenClp_setOption( givenClpPtr, 'f', BOOLEAN_TRUE );
			break;

		case '?':
			fprintf( stderr, "\n\nERROR: There is given either an option without an argument that\n" );
			fprintf( stderr, "requires one or an unknown option!\n" );
			fprintf( stderr, "Type 'retroflux -h' for a detailed help text!\n");
			AbortProgram;
			break;					
		}
	}
	
	/* testing whether the standard in is connected to the terminal; if so display help */
	if( isatty( STDIN_FILENO ) )
		displayHelpText( clpPtr );

	/* the '-o' argument is mandatory; if itisn't provided the program aborts with an appropriate message */
	if( GivenClp_getOption( givenClpPtr, 'o' ) == BOOLEAN_FALSE )
		MandatoryOption( "'-o'" );

	/* the '-r' argument is mandatory; if itisn't provided the program aborts with an appropriate message */
	if( GivenClp_getOption( givenClpPtr, 'r' ) == BOOLEAN_FALSE )
		MandatoryOption( "'-r'" );
	
	/* generating the name of the error log file and storing it in the structure CommandLineParameters */
	if( !( errorLogFile = calloc( strlen( CLP_getOutputFile( clpPtr ) ) + 5, sizeof( char ) ) ) )
		MemoryError( "parseClp", "errorLogFile" );
	strncpy( errorLogFile, CLP_getOutputFile( clpPtr ), strlen( CLP_getOutputFile( clpPtr ) ) + 1 );
	CLP_setErrorLogFile( clpPtr, strncat( errorLogFile, ".log", 4 ) );
	if( errorLogFile != NULL )
		free( errorLogFile );
	
	GivenClp_destroy( givenClpPtr ); /* destroying structure GivenCommandLineParameters */
}

/** Reads the data of a file and stores it in a structure @ref List.
* 
* @attention
* This is mainly a wrapper for function @ref readDataFromStream!
*
* @param filename the name of the file to read
* @param daylight_type indicates the file type (@ref DAYLIGHT_SMILES, @ref DAYLIGHT_SMIRKS, @ref DAYLIGHT_SMARTS)
* @param uniqueData if set to @ref BOOLEAN_TRUE input SMILES are filtered to yield unique SMILES
* @param errorLogFile file pointer on the error log file
* @retval List_Ptr points on structure @ref List containing the content of @c filename
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 09/06/2004 - UF - added new parameter @c uniqueData
*/
List_Ptr readDataFromFile( const char* const filename, const int daylight_type, \
	const int uniqueData, FILE* errorLogFile )
{
	FILE* dataFile = NULL; /* file pointer of the file with 'filename' */

	/* opening of the file and checking if this was successful */
	if( !( dataFile = fopen( filename, "r" ) ) )
		FileReadError( filename );
	
	/* calling of the sub-function that actually reads the file */
	return readDataFromStream( dataFile, filename, daylight_type, uniqueData, errorLogFile );
}

/** Reads the data of a stream and stores it in a structure @ref List.
*
* @param inputStream file pointer of the stream to read
* @param nameOfStream string indicating the name of the stream
* @param daylight_type indicates the file type (@ref DAYLIGHT_SMILES, @ref DAYLIGHT_SMIRKS, @ref DAYLIGHT_SMARTS)
* @param uniqueData if set to @ref BOOLEAN_TRUE input SMILES are filtered to yield unique SMILES
* @param errorLogFile file pointer on the error log file
* @retval List_Ptr points on structure @ref List containing the content of @c inputStream
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 08/12/2003 - Uli Fechner - considered unsuccessful creations of SmilesCompounds due to
* invalid input SMILES; added error output
* @version 09/06/2004 - UF - added new parameter @c uniqueData; bugfix: SMILES without a name were
* assigned an empty string;
* @version 11/06/2004 - UF - MAJOR change: from now on the first column is the SMILES and the second
* column is the name (if any; before the 1st column was the name and the 2nd column was the SMILES);
* this follows the common format of SMILES files
* @version 05/07/2004 - UF - if no name is provided the SMILES/SMARTS/SMIRKS is taken as a name
*/
List_Ptr readDataFromStream( FILE* inputStream, const char* const nameOfStream, \
	const int daylight_type, const int uniqueData, FILE* errorLogFile )
{
	int numberOfColumns = 0; /* the number of columns in the file provided via 'inputStream' */
	/* maximum number of characters per column of the file provided via 'inputStream' */
	int maxNumberOfCharsPerRow = 0;
	int numberOfRows = 0; /* the number of rows in the file provided via 'inputStream' */
	char* tempRow = NULL; /* temporary storage of a row of 'inputStream' */
	char* substring = NULL; /* pointer needed by the algorithm that divides a line into columns */
	char* tempSmiles = NULL; /* temporary storage for the smiles of the current row */
	char* tempName = NULL; /* temporary storage for the name of the current row */
	DoubleArrayPtr dAPtr = NULL; /* pointer on DoubleArray containing the results of getFileProperties */
	List_Ptr listPtr = NULL; /* pointer on DoubleLinkedList containing the content of the file */
	SmilesCompound_Ptr scPtr = NULL; /* temp. pointer on a SmilesCompound during insertion of a compound */
	char* errorMessage = NULL; /* for each unsuccessful compound creation a error message is created */
	/* pointer on structure stringArray to store the error messages for unsuccessful compound creations */
	StringArray_Ptr sAPtr = NULL;
	
	/* examine the file (number of columns, number of rows, maximum number of chars per row) */
	dAPtr = getFileProperties( inputStream, nameOfStream );
	
	rewind( inputStream );
	numberOfColumns = (int) DoubleArray_getValue( dAPtr, 0 );
	maxNumberOfCharsPerRow = (int) DoubleArray_getValue( dAPtr, 1 );
	numberOfRows = (int) DoubleArray_getValue( dAPtr, 2 );
	
	#if FLUX_DEBUG
	fprintf( stdout, "\nNoColumns = %d\tMaxNoChars = %d\tNoRows = %d\n", \
		numberOfColumns, maxNumberOfCharsPerRow, numberOfRows );
	#endif

	if( !( tempRow = calloc( maxNumberOfCharsPerRow + 2, sizeof( char ) ) ) )
		MemoryError( "tempRow", "readDataFromStream" );
	if( !( tempSmiles = calloc( maxNumberOfCharsPerRow + 2, sizeof( char ) ) ) )
		MemoryError( "tempSmiles", "readDataFromStream" );
	if( !( tempName = calloc( maxNumberOfCharsPerRow + 2, sizeof( char ) ) ) )
		MemoryError( "tempName", "readDataFromStream" );
	listPtr = List_create( uniqueData, SmilesCompound_display, SmilesCompound_destroy, \
		SmilesCompound_identical );
	sAPtr = StringArray_create( );

	/* in the following while loop each row of the file is read and stored in the strucure SmilesCompound */
	/* looping over all rows of 'inputSream' */
	while( fgets( tempRow, maxNumberOfCharsPerRow + 2, inputStream ) != NULL )
	{
		#if FLUX_DEBUG
		fprintf( stdout, "tempRow = %s", tempRow ); fflush( NULL );
		#endif

		/* the input file contains only one column - the SMILES string */
		if( numberOfColumns == 1 )
		{
			strncpy( tempSmiles, tempRow, strlen( tempRow ) - 1 );
			tempSmiles[ strlen( tempRow ) - 1 ] = '\0';
			strncpy( tempName, tempRow, strlen( tempRow ) - 1 );
			tempName[ strlen( tempRow ) - 1 ] = '\0';
		}
		/* the input file contains two columns - the name and the SMILES string */
		else
		{
			substring = strstr( tempRow, "\t" );
			strncpy( tempSmiles, tempRow, ( substring - tempRow ) );
			tempSmiles[ substring - tempRow ] = '\0';
			strncpy( tempName, substring + 1, ( tempRow + strlen( tempRow ) - 1 ) - ( substring + 1 ) );
			tempName[ ( tempRow + strlen( tempRow ) - 1 ) - ( substring + 1 ) ] = '\0';
		}
		#if FLUX_DEBUG
		fprintf( stdout, "Smiles: %s\tName: %s\n", tempSmiles, tempName ); fflush( NULL );
		#endif
		/* a SmilesCompound is created and inserted at the tail of the list
		if the insertion fails due to the presence of an identical SmilesCompound it is destroyed */
		scPtr = SmilesCompound_create( tempName, tempSmiles, daylight_type, errorLogFile );
		/* checking if the creation of the SmilesCompound was successful */
		if( scPtr != BOOLEAN_FALSE )
		{
			if( List_insertTail( listPtr, scPtr ) == BOOLEAN_FALSE )
				SmilesCompound_destroy( scPtr );
		}
		else
		{
			if( !( errorMessage = calloc( 300, sizeof( char ) ) ) )
				MemoryError( "errorMessage", "readDataFromStream" );
			sprintf( errorMessage, "Error reading compound %s\n", tempSmiles );
			StringArray_addElement( sAPtr, errorMessage );
		}
	}
	/* if there are any error messages they are printed to stderr */
	if( StringArray_getNumberOfElements( sAPtr ) != 0 )
		StringArray_display( sAPtr, stderr );
	
	/* cleaning up */
	free( tempRow );
	free( tempSmiles );
	free( tempName );
	DoubleArray_destroy( dAPtr );
	StringArray_destroy( sAPtr );
	
	return listPtr;
}

/** Examines the properties of a file stream.
* 
* The number of columns, characters per column and the number of rows are counted and the result
* of this examination are stored in a structure @ref DoubleArray. This @ref DoubleArray contains
* three elements regarded to the properties of dataFile:
*
* @li index 0 - number of columns
* @li index 1 - maximum number of characters per column
* @li index 2 - number of rows
*
* The file is also checked for integrity, i.e. it is checked, if it has the same number of columns in
* all rows. If this is not the case, the define @ref FileIntegrityError is called and the program
* aborts.
*
* @param inputStream file pointer of the stream to examine
* @param nameOfStream string indicating the name of the stream
* @retval DoubleArrayPtr contains the properties of @c inputStream
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
*/
DoubleArrayPtr getFileProperties( FILE* inputStream, const char* const nameOfStream )
{
	int numberOfColumns = 1; /* number of columns in dataFile */
	int maxCharsPerRow = 0; /* maximum number of characters per row in dataFile */
	int numberOfRows = 0; /* number of rows in dataFile */
	char currentCharacter; /* temporary storage of the character read by fgetc */
	int position; /* counter for the position in tempRow */
	char* tempRow; /* temporary string containing one row for analysis */
	char* substring; /* pointer needed for the 'number of occurences' algorithm */
	int counter = 1; /* counter needed by the 'number of occurences' algorithm */
	/* dAPtr points on a DoubleArray returning the results.
	The DoubleArray contains three values regarding to the properties of dataFile:
	index 0 - number of columns
	index 1 - maximum number of characters per column
	index 2 - number of rows */
	DoubleArrayPtr dAPtr; /* the file properties are stored in a DoubleArray structure */
	
	if( !( tempRow = calloc( 10000, sizeof( char ) ) ) )
		MemoryError( "tempRow", "getFileProperties" );	

	dAPtr = DoubleArray_create( 3 ); /* create a new DoubleArray with 3 elements */
	
	/* looping over all rows of the file */
	while( ( currentCharacter = fgetc( inputStream ) ) != EOF )
	{
		/* putting the character back in the stream */
		ungetc( currentCharacter, inputStream );
		
		/* resetting the position counter and the tab counter */
		position = 0;
		counter = 1;
		
		/* getting a line of inputStream */
		while( ( currentCharacter = fgetc( inputStream ) ) != '\n' )
		{
			tempRow[ position ] = currentCharacter;
			position++;
		}
		tempRow[ position ] = '\0';
		numberOfRows++;
		
		/* counting the columns by counting the number of occurences of '\t' */
		substring = strstr( tempRow, "\t" );
		while( substring != NULL)
		{
			counter++;			
			substring = strstr( substring + 1, "\t" );
		}
		
		/* if the number of columns of the first row are counted the variable 'numberOfColumns' is set */
		if( numberOfRows == 1 )
			numberOfColumns = counter;
		
		/* if the number of columns in one row is different from that in the first row the program aborts */
		if( counter != numberOfColumns )			
			FileIntegrityError( nameOfStream, numberOfRows, counter, numberOfColumns );
			
		/* keeping the maximum number of chars in one row up to date */
		if( strlen( tempRow ) > maxCharsPerRow )
			maxCharsPerRow = position;
	}

	/* checking if there are exactly 2 columns; if not the input file is invalid and the program aborts */
	if( numberOfColumns > 2 )
	{
		fprintf( stderr, "\nERROR: The number of columns in %s is neither one nor two!\n", nameOfStream );
		fprintf( stderr, "Only input files with the name in the first column and the SMILES string\n" );
		fprintf( stderr, "in the second column or the SMILES string as the only column are valid!\n" );
		AbortProgram;
	}
	
	/* the values regarding to the file property are copied to the struct DoubleArray */
	DoubleArray_setValue( dAPtr, 0, numberOfColumns );
	DoubleArray_setValue( dAPtr, 1, maxCharsPerRow );
	DoubleArray_setValue( dAPtr, 2, numberOfRows );
	
	free( tempRow );
	
	return dAPtr;
}

/** Displays the help text on standard error.
*
* @param clpPtr pointer on structure @ref CommandLineParameters that is used to store all the command line parameters
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
* @version 7/1/2003 - UF - added help text for the f option
*/
void displayHelpText( const CLP_Ptr clpPtr )
{
	fprintf( stderr, "\nNAME:\n" );
  fprintf( stderr, "  retroflux\n" );
	fprintf( stderr, "\nFUNCTION:\n" );
  fprintf( stderr, "  Facilitates a virtual retro-synthesis of compounds.\n" );
	fprintf( stderr, "\nUSAGE:\n" );
	fprintf( stderr, "  retroflux [Options] <INFILE\n" );
	fprintf( stderr, "     INFILE has to be a file with one SMILES per row or a tab separated file\n" );
	fprintf( stderr, "     with SMILES in the first and labels in the second column.\n" );
	fprintf( stderr, "\nREMARK:\n" );
	fprintf( stderr, "  An error log file is automatically created. Its name is the suffix '.log'\n" );
	fprintf( stderr, "  appended to the name of the output file.\n" );
	fprintf( stderr, "\nOPTIONS:\n" );
	fprintf( stderr, "  -h\n" );
	fprintf( stderr, "     Display this help text.\n" );
	fprintf( stderr, "  -v\n" );
	fprintf( stderr, "     Display detailed version information and exit.\n" );
	fprintf( stderr, "  -o STRING\n" );
	fprintf( stderr, "     Set STRING as the name of the output file.\n" );
	fprintf( stderr, "  -r FILENAME | mandatory\n" );
	fprintf( stderr, "     Set FILENAME as the name of the reaction SMILES file. This file has to\n" );
	fprintf( stderr, "     contain exactly two tab separated columns, where SMIRKS are in the first\n" );
	fprintf( stderr, "     column and reaction names are in the second column.\n" );
	fprintf( stderr, "  -f FILENAME\n" );
	fprintf( stderr, "     Set FILENAME as the name of the filter SMILES file. This file has to\n" );
	fprintf( stderr, "     contain one SMILES per row. All molecules in the filter file are\n" );
	fprintf( stderr, "     considered as unwanted fragments, i.e. reactions that lead to one of these\n" );
	fprintf( stderr, "     molecules are not carried out.\n" );
	fprintf( stderr, "     If no file is specified, all resulting fragments are considered as valid.\n" );
	fprintf( stderr, "\nAUTHOR:\n" );
	fprintf( stderr, "  Uli Fechner\n" );
	fprintf( stderr, "\nVERSION & RELEASE DATE:\n" );
  fprintf( stderr, "  %s ( %s )\n", RETROFLUX_VERSION, RETROFLUX_DATE );
	fprintf( stderr, "\nBUGS:\n" );
  fprintf( stderr, "  Please report bugs to u.fechner@chemie.uni-frankfurt.de\n\n" );
	exit( EXIT_SUCCESS );
}

/** Displays the version information text on standard error.
*
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
*/
void displayVersionInformation( void )
{
	fprintf( stderr, "\nVERSION & RELEASE DATE:\n\n");
	fprintf( stderr, "  includes.h: %s [%s]\n", INCLUDES_VERSION, INCLUDES_DATE );
	fprintf( stderr, "  generalDefines.h: %s [%s]\n", GENERALDEFINES_VERSION, GENERALDEFINES_DATE );
	fprintf( stderr, "  givenClp.c: %s [%s]\n", GIVENCLP_VERSION, GIVENCLP_DATE );
	fprintf( stderr, "  doubleArray.c: %s [%s]\n", DOUBLEARRAY_VERSION, DOUBLEARRAY_DATE );
	fprintf( stderr, "  clp.c: %s [%s]\n", CLP_VERSION, CLP_DATE );
	fprintf( stderr, "  smilesCompound.c: %s [%s]\n", SMILESCOMPOUND_VERSION, SMILESCOMPOUND_DATE );
	fprintf( stderr, "  doubleLinkedList.c: %s [%s]\n", DOUBLELINKEDLIST_VERSION, DOUBLELINKEDLIST_DATE );
	fprintf( stderr, "  retroflux core: %s [%s]\n\n", RETROFLUX_VERSION, RETROFLUX_DATE );
	exit( EXIT_SUCCESS );
}
